import logging
import warnings
from pathlib import Path

import pandas as pd

here = Path(__file__).parent
logger = logging.getLogger(__name__)


def deprecation_warning(msg, item="This method", stacklevel=3):
    """Write deprecation warning to log"""
    warnings.simplefilter("always", DeprecationWarning)
    message = f"{item} is deprecated and will be removed in future versions. {msg}"
    warnings.warn(message, DeprecationWarning, stacklevel=stacklevel)


def raise_data_error(msg, data):
    """Format error message with (head of) data table and raise"""
    raise ValueError(format_log_message(msg, data))


def format_log_message(msg, data):
    """Utils function to format message with (head of) data table"""
    if isinstance(data, pd.MultiIndex):
        data = data.to_frame(index=False)
    data = data.drop_duplicates()
    return f"{msg}:\n{data.head()}" + ("\n..." if len(data) > 5 else "")



import numpy as np
import pandas as pd


def _compare(
    left, right, left_label="left", right_label="right", drop_close=True, **kwargs
):
    """Internal implementation of comparison of IamDataFrames or pd.Series"""

    def as_series(s):
        return s if isinstance(s, pd.Series) else s._data

    ret = pd.merge(
        left=as_series(left).rename(index=left_label),
        right=as_series(right).rename(index=right_label),
        how="outer",
        left_index=True,
        right_index=True,
    )
    if drop_close:
        ret = ret[~np.isclose(ret[left_label], ret[right_label], **kwargs)]
    return ret



import re

import numpy as np
import pandas as pd
from pandas.api.types import is_list_like

REGEXP_CHARACTERS = r".^$+?()[]{}|"


def concat_with_pipe(x, *args, cols=None):
    """Concatenate a list or pandas.Series using ``|``, drop None or numpy.nan"""
    if args:
        # Guard against legacy-errors when adding `*args` (#778)
        # TODO: deprecated, remove for release >= 3.1
        for i in args:
            if is_list_like(i):
                raise DeprecationWarning(f"Please use `cols={i}`.")
        x = [x] + list(args)
    cols = cols or (x.index if isinstance(x, pd.Series) else range(len(x)))
    return "|".join([x[i] for i in cols if x[i] not in [None, np.nan, ""]])


def find_depth(data, s="", level=None):
    """Return or assert the depth (number of ``|``) of variables

    Parameters
    ----------
    data : str or list of strings
        IAMC-style variables
    s : str, default ''
        remove leading `s` from any variable in `data`
    level : int or str, optional
        If None, return depth (number of ``|``); else, return list of booleans
        whether depth satisfies the condition (equality if level is int,
        >= if ``.+``,  <= if ``.-``)
    """
    if is_list_like(level):
        raise ValueError(
            "Level is only run with ints or strings, not lists. Use strings with "
            "integers and + or - to filter by ranges."
        )
    if is_str(data):
        return _find_depth([data], s, level)[0]

    return _find_depth(data, s, level)


def _find_depth(data, s="", level=None):
    """Internal implementation of `find_depth()´"""
    # remove wildcard as last character from string, escape regex characters
    _s = re.compile("^" + escape_regexp(s.rstrip("*")))
    _p = re.compile("\\|")

    # find depth
    def _count_pipes(val):
        return len(_p.findall(re.sub(_s, "", val))) if _s.match(val) else None

    n_pipes = map(_count_pipes, data if is_list_like(data) else list(data))

    # if no level test is specified, return the depth as (list of) int
    if level is None:
        return list(n_pipes)

    # if `level` is given, set function for finding depth level =, >=, <= |s
    if not is_str(level):
        # test = lambda x: level == x if x is not None else False
        def test(x):
            return level == x if x is not None else False

    elif level[-1] == "-":
        level = int(level[:-1])

        # test = lambda x: level >= x if x is not None else False
        def test(x):
            return level >= x if x is not None else False

    elif level[-1] == "+":
        level = int(level[:-1])

        # test = lambda x: level <= x if x is not None else False
        def test(x):
            return level <= x if x is not None else False

    else:
        raise ValueError(f"Unknown level type: `{level}`")

    return list(map(test, n_pipes))


def get_variable_components(x, level, join=False):
    """Return components for requested level in a list or join these in a str.

    Parameters
    ----------
    x : str
        Uses ``|`` to separate the components of the variable.
    level : int or list of int
        Position of the component.
    join : bool or str, optional
        If True, IAMC-style (``|``) is used as separator for joined components.

    Returns
    -------
    str
    """
    _x = x.split("|")
    if join is False:
        return [_x[i] for i in level] if is_list_like(level) else _x[level]
    else:
        level = [level] if isinstance(level, int) else level
        join = "|" if join is True else join
        return join.join([_x[i] for i in level])


def reduce_hierarchy(x, depth):
    """Reduce the hierarchy (indicated by ``|``) of x to the specified depth

    Parameters
    ----------
    x : str
        Uses ``|`` to separate the components of the variable.
    depth : int or list of int
        Position of the components.

    """
    _x = x.split("|")
    depth = len(_x) + depth - 1 if depth < 0 else depth
    return "|".join(_x[0 : (depth + 1)])


def escape_regexp(s):
    """Escape characters with specific regexp use"""
    s = str(s)
    for c in REGEXP_CHARACTERS:
        s = s.replace(c, "\\" + c)
    # pyam uses `*` as wildcard, replace with `.*` for regex
    s = s.replace("*", ".*")
    return s


def is_str(x):
    """Returns True if x is a string"""
    return isinstance(x, str)



import logging
from itertools import compress

import numpy as np
import pandas as pd

from pyam._compare import _compare
from pyam.exceptions import format_log_message
from pyam.index import replace_index_values
from pyam.str import find_depth, is_str, reduce_hierarchy
from pyam.utils import adjust_log_level, is_list_like, to_list

logger = logging.getLogger(__name__)


def _aggregate(df, variable, components=None, method="sum"):
    """Internal implementation of the `aggregate` function"""

    if components is not None:
        # ensure that components is a proper list (not a dictionary)
        if not is_list_like(components) or isinstance(components, dict):
            raise ValueError(
                f"Value for `components` must be a list, found: {components}"
            )

        # list of variables require default components (no manual list)
        if is_list_like(variable):
            raise NotImplementedError(
                "Aggregating by list of variables does not support `components`."
            )

    mapping = {}
    msg = "Cannot aggregate variable '{}' because it has no components."
    # if single variable
    if is_str(variable):
        # default components to all variables one level below `variable`
        components = components or df._variable_components(variable)

        if not len(components):
            logger.warning(msg.format(variable))
            return

        for c in components:
            mapping[c] = variable

    # else, use all variables one level below `variable` as components
    else:
        for v in variable if is_list_like(variable) else [variable]:
            _components = df._variable_components(v)
            if not len(_components):
                logger.warning(msg.format(v))
                continue

            for c in _components:
                mapping[c] = v

    # rename all components to `variable` and aggregate
    _df = df._data[df._apply_filters(variable=mapping.keys())]
    _df.index = replace_index_values(_df, "variable", mapping)
    return _group_and_agg(_df, [], method)


def _aggregate_recursive(df, variable, recursive):
    """Recursive aggregation along the variable tree"""

    # downselect to components of `variable`, initialize list for aggregated (new) data
    # keep variable at highest level if it exists
    _df = df.filter(variable=[variable, f"{variable}|*"])
    data_list = []

    # iterate over variables (bottom-up) and aggregate all components up to `variable`
    for d in reversed(range(find_depth(variable), max(find_depth(_df.variable)))):
        components = compress(_df.variable, find_depth(_df.variable, level=d + 1))
        var_list = {reduce_hierarchy(v, -1) for v in components}

        # a temporary dataframe allows to distinguish between full data and new data
        _data_agg = _aggregate(_df, variable=var_list)

        # check if data for intermediate variables already exists
        with adjust_log_level("pyam.core"):
            _data_self = _df.filter(variable=var_list)._data
        _overlap = _data_agg.index.intersection(_data_self.index)
        _new = _data_agg.index.difference(_data_self.index)

        # assert that aggregated values are consistent with existing data (optional)
        if recursive != "skip-validate" and not _overlap.empty:
            conflict = _compare(_data_self, _data_agg[_overlap], "self", "aggregate")
            if not conflict.empty:
                msg = "Aggregated values are inconsistent with existing data:"
                raise ValueError(f"{msg}\n{conflict}")

        # append aggregated values that are not already in data
        _df.append(_data_agg[_new], inplace=True)
        data_list.append(_data_agg[_new])

    return pd.concat(data_list)


def _aggregate_region(
    df,
    variable,
    region,
    subregions=None,
    components=False,
    method="sum",
    weight=None,
    drop_negative_weights=True,
):
    """Internal implementation for aggregating data over subregions"""
    if not is_str(variable) and components is not False:
        raise ValueError(
            "Aggregating by list of variables with components is not supported."
        )

    if weight is not None and components is not False:
        raise ValueError("Using weights and components in one operation not supported.")

    # default subregions to all regions other than `region`
    if weight is None:
        subregions = subregions or df._all_other_regions(region, variable)
    else:
        subregions = subregions or df._all_other_regions(region, [variable, weight])

    if not len(subregions):
        logger.warning(
            f"Cannot aggregate variable '{variable}' to '{region}' "
            "because it does not exist in any subregion."
        )
        return

    # compute aggregate over all subregions
    subregion_df = df.filter(region=subregions)
    rows = subregion_df._apply_filters(variable=variable)
    if weight is None:
        if drop_negative_weights is False:
            raise ValueError(
                "Dropping negative weights can only be used with `weights`."
            )

        _data = _group_and_agg(subregion_df._data[rows], "region", method=method)
    else:
        weight_rows = subregion_df._apply_filters(variable=weight)
        _data = _agg_weight(
            subregion_df._data[rows],
            subregion_df._data[weight_rows],
            method,
            drop_negative_weights,
        )

    # if not `components=False`, add components at the `region` level
    if components:
        with adjust_log_level("pyam.core"):
            region_df = df.filter(region=region)

        # if `True`, auto-detect `components` at the `region` level,
        # defaults to variables below `variable` only present in `region`
        if components is True:
            level = dict(level=None)
            r_comps = region_df._variable_components(variable, **level)
            sr_comps = subregion_df._variable_components(variable, **level)
            components = set(r_comps).difference(sr_comps)

        if len(components):
            # rename all components to `variable` and aggregate
            rows = region_df._apply_filters(variable=components)
            _df = region_df._data[rows]
            mapping = {c: variable for c in components}
            _df.index = replace_index_values(_df.index, "variable", mapping)
            _data = _data.add(_group_and_agg(_df, "region"), fill_value=0)

    return _data.dropna()


def _aggregate_time(df, variable, column, value, components, method="sum"):
    """Internal implementation for aggregating data over subannual time"""
    # default `components` to all entries in `column` other than `value`
    if components is None:
        components = list(set(df.data.subannual.unique()) - {value})

    # compute aggregate over time
    filter_args = dict(variable=variable)
    filter_args[column] = components
    index = [d for d in df.dimensions if d != column]

    _data = pd.concat(
        [
            df.filter(**filter_args)
            .data.pivot_table(index=index, columns=column)
            .value.rename_axis(None, axis=1)
            .apply(method, axis=1)
        ],
        names=[column] + index,
        keys=[value],
    )

    # reset index-level order to original IamDataFrame
    _data.index = _data.index.reorder_levels(df.dimensions)

    # set name of series
    _data.name = "value"

    return _data


def _group_and_agg(df, by, method="sum"):
    """Group-by & aggregate `pd.Series` by index names on `by`"""
    cols = df.index.names.difference(to_list(by))
    # pick aggregator func (default: sum)
    return df.groupby(cols).agg(method)


def _agg_weight(data, weight, method, drop_negative_weights):
    """Aggregate `data` by regions with weights, return indexed `pd.Series`"""

    # only summation allowed with weights
    if method not in ["sum", np.sum]:
        raise ValueError("Only method 'np.sum' allowed for weighted average.")

    weight = weight.droplevel(["variable", "unit"])
    data_index = data.droplevel(["variable", "unit"]).index

    # check that weights exist for all data rows
    missing_weights = data_index.difference(weight.index)
    if not missing_weights.empty:
        raise ValueError(
            format_log_message(
                "Missing weights for the following data rows", missing_weights
            )
        )

    # warn if no data exists for available weights
    missing_data = weight.index.difference(data_index)
    if not missing_data.empty:
        logger.warning(
            format_log_message(
                "Ignoring weights for the following missing data rows", missing_data
            )
        )
        weight[missing_data] = np.nan

    # remove (and warn) negative values from weights due to strange behavior
    if drop_negative_weights is True:
        if any(weight < 0):
            logger.warning(
                "Some weights are negative. Data weighted by negative values will be "
                "dropped. To use both positive and negative weights, "
                "please use the keyword argument `drop_negative_weights=False`."
            )
            # drop negative weights
            weight[weight < 0] = None

    col1 = data.index.names.difference(["region"])
    col2 = data.index.names.difference(["region", "variable", "unit"])
    return (data * weight).groupby(col1).apply(
        pd.Series.sum, skipna=False
    ) / weight.groupby(col2).sum()



import dateutil
import pandas as pd

from pyam.exceptions import raise_data_error
from pyam.index import append_index_col


def swap_time_for_year(df, inplace=False, subannual=False):
    """Internal implementation to swap 'time' domain to 'year' (as int)"""
    if not df.time_col == "time":
        raise ValueError("Time domain must be datetime to use this method")

    ret = df.copy() if not inplace else df

    index = ret._data.index
    time = pd.Series(index.get_level_values("time"))
    order = [v if v != "time" else "year" for v in index.names]

    # reduce "time" index column to "year"
    # TODO use `replace_index_values` instead of `append_index_col`
    index = index.droplevel("time")
    new_index_col = time.apply(lambda x: x if isinstance(x, int) else x.year)
    index = append_index_col(index, new_index_col, "year", order=order)

    # if selected, extract the "subannual" info from the "time" index column
    if subannual:
        # if subannual is True, default to simple datetime format without year
        if subannual is True:
            subannual = "%m-%d %H:%M%z"
        if isinstance(subannual, str):
            _subannual = time.apply(lambda x: x.strftime(subannual))
        else:
            _subannual = time.apply(subannual)

        index = append_index_col(index, _subannual, "subannual")
        ret.extra_cols.append("subannual")

    rows = index.duplicated()
    if any(rows):
        error_msg = "Swapping time for year causes duplicates in `data`"
        raise_data_error(error_msg, index[rows].to_frame().reset_index(drop=True))

    # assign data and other attributes
    ret._data.index = index
    ret.time_col = "year"
    ret._set_attributes()

    if not inplace:
        return ret


def swap_year_for_time(df, inplace=False):
    """Internal implementation to swap 'year' domain to 'time' (as datetime)"""

    if not df.time_col == "year":
        raise ValueError("Time domain must be 'year' to use this method")

    ret = df.copy() if not inplace else df
    index = ret._data.index

    order = [v if v != "year" else "time" for v in index.names]

    if "subannual" in df.extra_cols:
        order = order.remove("subannual")
        time_values = zip(*[index.get_level_values(c) for c in ["year", "subannual"]])
        time = list(map(dateutil.parser.parse, [f"{y}-{s}" for y, s in time_values]))
        index = index.droplevel(["year", "subannual"])
        ret.extra_cols.remove("subannual")
    else:
        time = index.get_level_values("year")
        index = index.droplevel(["year"])

    # add new index column, assign data and other attributes
    index = append_index_col(index, time, "time", order=order)
    ret._data.index = index
    ret.time_col = "time"
    ret._set_attributes()
    delattr(ret, "year")

    if not inplace:
        return ret



from copy import deepcopy

import numpy as np
import pandas as pd

from pyam import filter_by_meta
from pyam.str import is_str
from pyam.utils import META_IDX, is_list_like


class Statistics:
    """This class generates descriptive statistics of timeseries data

    Parameters
    ----------
    df : IamDataFrame
        an IamDataFrame from which to retrieve meta indicators for grouping
        or filtering
    groupby : str or dict
        a column of `df.meta` to be used for `groupby`
        or a dictionary of `{column: list}`, where `list` is used for ordering
    filters : list of tuples
        arguments for filtering and describing, either `((index, dict)` or
        `((index[0], index[1]), dict)`; when also using `groupby`, index must
        have length 2.
    percentiles : list-like of numbers, optional
        The percentiles to get from :meth:`pandas.DataFrame.describe()`.
        All should fall between 0 and 1. The default is `[.25, .5, .75]`,
        which returns the 25th, 50th, and 75th percentiles.
    """

    def __init__(  # noqa: C901
        self, df, groupby=None, filters=None, rows=False, percentiles=[0.25, 0.5, 0.75]
    ):
        self.df = df
        self.idx_depth = None

        # assing `groupby` settings and check that specifications are valid
        self.col = None
        self.groupby = None
        if is_str(groupby):
            self.col = groupby
            self.groupby = {groupby: None}
        elif isinstance(groupby, dict) and len(groupby) == 1:
            self.col = list(groupby.keys())[0]
            self.groupby = groupby
            self.idx_depth = 2
        elif groupby is not None:
            raise ValueError(f"arg `{groupby}` not valid `groupby`")
        if self.col is not None and self.col not in df.meta.columns:
            raise ValueError(f"column `{self.col}` not in `df.meta`")

        # if neither groupby nor filters is given, use filters to describe all
        # and assume that rows are used
        if groupby is None and filters is None:
            self.filters = [("", {})]
            rows = True
        else:
            self.filters = filters if filters is not None else []

        # set lists to sort index and subindex
        self._idx = [] if self.col is None else [self.col]
        self._sub_idx = (
            self.groupby[self.col] or self.df[self.col].unique()
            if self.col is not None
            else []
        )
        self._headers, self._subheaders = ([], [])

        # assing `filters` settings and check that specifications are valid
        for idx, _filter in self.filters:
            # check that index in tuple is valid
            if is_str(idx):
                self._add_to_index(idx)
            else:
                if not (
                    isinstance(idx, tuple)
                    and len(idx) == 2
                    and is_str(idx[0])
                    or not is_str(idx[1])
                ):
                    raise ValueError(f"`{idx}` is not a valid index")
                self._add_to_index(idx[0], idx[1])
            # check that filters in tuple are valid
            if not isinstance(_filter, dict):
                raise ValueError(f"`{_filter}` is not a valid filter")
            elif not (set(_filter) - set(META_IDX)).issubset(df.meta):
                raise ValueError(
                    "column `{}` not in `df.meta`".format(
                        set(_filter) - set(META_IDX) - set(df.meta)
                    )
                )

        self.stats = None
        self.rows = [] if rows else None

        # percentiles for passing to `pandas.describe()`
        self.percentiles = list(percentiles)
        self._describe_cols = (
            ["count", "mean", "std", "min"]
            + [f"{i:.0%}" for i in self.percentiles]
            + ["max"]
        )

    def _add_to_index(self, idx, sub_idx=None):
        # assign index depth if not set
        if self.idx_depth is None:
            self.idx_depth = 1 if sub_idx is None else 2
        # check that index matches depth
        if self.groupby is not None and sub_idx is None:
            msg = "if `groupby` is used, index `{}` must have format `{}`"
            raise ValueError(msg.format(idx, "(idx0, idx1)"))
        if self.idx_depth == 1 and sub_idx is not None:
            raise ValueError(f"index depth set to 1, found `({idx}, {sub_idx})`")
        if self.idx_depth == 2 and sub_idx is None:
            raise ValueError(f"index depth set to 2, found `({idx})`")

        # append to lists for sorting index
        if idx not in self._idx:
            self._idx.append(idx)
        if self.idx_depth == 2 and sub_idx not in self._sub_idx:
            self._sub_idx.append(sub_idx)

    def _add_to_header(self, header, subheader):
        if header not in self._headers:
            self._headers.append(header)
        if is_list_like(subheader):
            for s in subheader:
                if s not in self._subheaders:
                    self._subheaders.append(s)
        elif subheader not in self._subheaders:
            self._subheaders.append(subheader)

    def add(self, data, header, row=None, subheader=None):  # noqa: C901
        """Filter 'data' by arguments of this Statistics instance,

        Apply :meth:`pandas.DataFrame.describe()` and format the statistics

        Parameters
        ----------
        data : pandas.DataFrame or pandas.Series
            data for which summary statistics should be computed
        header : str
            column name for descriptive statistics
        row : str
            row name for descriptive statistics
            (required if :class:`Statistics(rows=True) <Statistics>`)
        subheader : str, optional
            column name (level=1) if data is a unnamed :class:`pandas.Series`
        """
        # verify validity of specifications
        if self.rows is not None and row is None:
            raise ValueError("row specification required")
        if self.rows is None and row is not None:
            raise ValueError("row arg illegal for this `Statistics` instance")
        if isinstance(data, pd.Series):
            if subheader is not None:
                data.name = subheader
            elif data.name is None:
                msg = "`data` must be named `pd.Series` or provide `subheader`"
                raise ValueError(msg)
            data = pd.DataFrame(data)

        if self.rows is not None and row not in self.rows:
            self.rows.append(row)

        _stats = None

        # describe with groupby feature
        if self.groupby is not None:
            filter_args = dict(data=data, df=self.df, join_meta=True)
            filter_args.update(self.groupby)
            _stats = (
                filter_by_meta(**filter_args)
                .groupby(self.col)
                .describe(percentiles=self.percentiles)
            )
            _stats = pd.concat([_stats], keys=[self.col], names=[""], axis=0)
            if self.rows:
                _stats["row"] = row
                _stats.set_index("row", append=True, inplace=True)
            _stats.index.names = [""] * 3 if self.rows else [""] * 2

        # describe with filter feature
        for idx, _filter in self.filters:
            filter_args = dict(data=data, df=self.df)
            filter_args.update(_filter)
            _stats_f = filter_by_meta(**filter_args).describe(
                percentiles=self.percentiles
            )
            _stats_f = pd.DataFrame(_stats_f.unstack()).T
            if self.idx_depth == 1:
                levels = [[idx]]
            else:
                levels = [[idx[0]], [idx[1]]]
            lvls, lbls = (
                (levels, [[0]] * self.idx_depth)
                if not self.rows
                else (levels + [[row]], [[0]] * (self.idx_depth + 1))
            )
            _stats_f.index = pd.MultiIndex(levels=lvls, codes=lbls)
            _stats = _stats_f if _stats is None else pd.concat([_stats, _stats_f])

        # add header
        _stats = pd.concat([_stats], keys=[header], names=[""], axis=1)
        _stats.index.names = [None] * len(_stats.index.names)
        subheader = _stats.columns.get_level_values(1).unique()
        self._add_to_header(header, subheader)

        # set statistics
        if self.stats is None:
            self.stats = _stats
        else:
            self.stats = _stats.combine_first(self.stats)

    def reindex(self, copy=True):
        """Reindex the summary statistics dataframe"""
        ret = deepcopy(self) if copy else self

        ret.stats = ret.stats.reindex(index=ret._idx, level=0)
        if ret.idx_depth == 2:
            ret.stats = ret.stats.reindex(index=ret._sub_idx, level=1)
        if ret.rows is not None:
            ret.stats = ret.stats.reindex(index=ret.rows, level=ret.idx_depth)

        ret.stats = ret.stats.reindex(columns=ret._headers, level=0)
        ret.stats = ret.stats.reindex(columns=ret._subheaders, level=1)
        ret.stats = ret.stats.reindex(columns=ret._describe_cols, level=2)

        if copy:
            return ret

    def summarize(
        self, center="mean", fullrange=None, interquartile=None, custom_format="{:.2f}"
    ):
        """Format the compiled statistics to a concise string output

        Parameters
        ----------
        center : str, default `mean`
            what to return as 'center' of the summary: `mean`, `50%`, `median`
        fullrange : bool, default None
            return full range of data if True or `fullrange`, `interquartile`
            and `format_spec` are None
        interquartile : bool, default None
            return interquartile range if True
        custom_format : formatting specifications
        """
        # call `reindex()` to reorder index and columns
        self.reindex(copy=False)

        center = "median" if center == "50%" else center
        if fullrange is None and interquartile is None:
            fullrange = True
        return self.stats.apply(
            format_rows,
            center=center,
            fullrange=fullrange,
            interquartile=interquartile,
            custom_format=custom_format,
            axis=1,
            raw=False,
        )


# %% auxiliary functions


def format_rows(
    row, center, fullrange=None, interquartile=None, custom_format="{:.2f}"
):
    """Format a row with `describe()` columns to a concise string"""
    if (fullrange or 0) + (interquartile or 0) == 1:
        legend = "{} ({})".format(
            center, "max, min" if fullrange is True else "interquartile range"
        )

        row_index = row.index.droplevel(2).drop_duplicates()
        ret_index = pd.MultiIndex.from_tuples([("count", "")]).append(row_index)
        ret_index.names = [None, legend]
    else:
        raise ValueError("Use either fullrange or interquartile range.")

    ret = pd.Series(index=ret_index, dtype=float)

    row = row.sort_index()
    center = "50%" if center == "median" else center

    # get maximum of `count` and write to first entry of return series
    count = max(
        [i for i in row.loc[(slice(None), slice(None), "count")] if not np.isnan(i)]
    )
    ret.loc[("count", "")] = (f"{count:.0f}") if count > 1 else ""

    # set upper and lower for the range
    upper, lower = ("max", "min") if fullrange is True else ("75%", "25%")

    # format `describe()` columns to string output
    for i in row_index:
        x = row.loc[i]
        _count = x["count"]
        if np.isnan(_count) or _count == 0:
            s = ""
        elif _count > 1:
            s = "{f} ({f}, {f})".format(f=custom_format).format(
                x[center], x[upper], x[lower]
            )
        elif _count == 1:
            s = f"{custom_format}".format(x["50%"])
        # add count of this section as `[]` if different from count_max
        if 0 < _count < count:
            s += f" [{_count:.0f}]"
        ret.loc[i] = s

    return ret



import logging
import sys
from importlib.metadata import PackageNotFoundError, version
from pathlib import Path

import yaml

from pyam.core import (
    IamDataFrame,
    compare,
    concat,
    filter_by_meta,
    read_datapackage,
)
from pyam.iiasa import lazy_read_iiasa, read_iiasa
from pyam.ixmp4 import read_ixmp4
from pyam.netcdf import read_netcdf
from pyam.run_control import run_control
from pyam.statistics import Statistics
from pyam.testing import assert_iamframe_equal
from pyam.unfccc import read_unfccc
from pyam.utils import IAMC_IDX
from pyam.worldbank import read_worldbank

here = Path(__file__).parent

try:
    __IPYTHON__  # type: ignore
    _in_ipython_session = True
except NameError:
    _in_ipython_session = False

_sys_has_ps1 = hasattr(sys, "ps1")

# Logging is only configured by default when used in an interactive environment.
# This follows the setup in ixmp4 and nomenclature.
if _in_ipython_session or _sys_has_ps1:
    with open(here / "logging.yaml") as file:
        logging.config.dictConfig(yaml.safe_load(file))

try:
    __version__ = version("pyam-iamc")
# the pyam package is distributed under different names on pypi and conda
except PackageNotFoundError:
    __version__ = version("pyam")



import logging
import math

import numpy as np
import pandas as pd

from pyam.str import is_str
from pyam.utils import raise_data_error, to_int

logger = logging.getLogger(__name__)


def fill_series(x, time):
    """Returns the timeseries value at a point in time by linear interpolation

    Parameters
    ----------
    x : pandas.Series
        a timeseries to be interpolated
    time : int or pandas.datetime
        year or datetime to interpolate
    """
    x = x.dropna()
    if time in x.index and not np.isnan(x[time]):
        return x[time]
    else:
        prev = [i for i in x.index if i < time]
        nxt = [i for i in x.index if i > time]
        if prev and nxt:
            p = max(prev)
            n = min(nxt)
            return ((n - time) * x[p] + (time - p) * x[n]) / (n - p)
        else:
            return np.nan


def cumulative(x, first_year, last_year):
    """Returns the cumulative sum of a timeseries

    This function implements linear interpolation between years
    and ignores nan's in the range.
    The function includes the last-year value of the series, and
    raises a warning if start_year or last_year is outside of
    the timeseries range and returns nan

    Parameters
    ----------
    x : pandas.Series
        a timeseries to be summed over time
    first_year : int
        first year of the sum
    last_year : int
        last year of the sum (inclusive)
    """
    # if the timeseries does not cover the range `[first_year, last_year]`,
    # return nan to avoid erroneous aggregation
    if min(x.index) > first_year:
        logger.warning(f"Start of period {first_year} outside of range.")
        return np.nan
    if max(x.index) < last_year:
        logger.warning(f"End of period {last_year} outside of range.")
        return np.nan

    # make sure we're using integers
    to_int(x, index=True)

    x[first_year] = fill_series(x, first_year)
    x[last_year] = fill_series(x, last_year)

    years = [
        i for i in x.index if i >= first_year and i <= last_year and ~np.isnan(x[i])
    ]
    years.sort()

    # loop over years
    if not np.isnan(x[first_year]) and not np.isnan(x[last_year]):
        value = 0
        for i, yr in enumerate(years[:-1]):
            next_yr = years[i + 1]
            # the summation is shifted to include the first year fully in sum,
            # otherwise, would return a weighted average of `yr` and `next_yr`
            value += ((next_yr - yr - 1) * x[next_yr] + (next_yr - yr + 1) * x[yr]) / 2

        # the loop above does not include the last element in range
        # (`last_year`), therefore added explicitly
        value += x[last_year]

        return value


def cross_threshold(
    x, threshold=0, direction=["from above", "from below"], return_type=int
):
    """Returns a list of the years in which a timeseries crosses a threshold

    Parameters
    ----------
    x : :class:`pandas.Series`
        A timeseries indexed over years (as integers)
    threshold : float, optional
        The threshold that the timeseries is checked against
    direction : str, optional
        Whether to return all years where the threshold is crossed
        or only where threshold is crossed in a specific direction
    return_type : type, optional
        Whether to cast the returned values to integer (years)
    """
    direction = [direction] if is_str(direction) else list(direction)
    if not set(direction).issubset({"from above", "from below"}):
        raise ValueError(f"Invalid direction: {direction}")

    # get the values and time-domain index
    x = x.dropna()
    values, index = x.values - threshold, x.index.to_numpy()
    positive, negative = (values >= 0), (values < 0)

    # determine all indices before crossing the threshold
    pre = [False] * (len(x) - 1)
    if "from above" in direction:
        pre |= positive[:-1] & negative[1:]
    if "from below" in direction:
        pre |= positive[1:] & negative[:-1]
    pre = np.argwhere(pre)
    # determine all indices after crossing the threshold
    post = pre + 1

    # compute the index value where the threshold is crossed
    change = (values[post] - values[pre]) / (index[post] - index[pre])
    years = index[pre] - values[pre] / change

    # it year (as int) is returned, add one because int() rounds down
    if return_type is int:
        return [y + 1 for y in map(int, years)]
    return years


def growth_rate(x):
    """Compute the annualized growth rate from timeseries data

    The annualized growth rate parameter in period *t* is computed assuming exponential
    growth based on the changes from period *t* to period *t+1*.

    Parameters
    ----------
    x : :class:`pandas.Series`
        Timeseries data indexed over the time domain.

    Returns
    -------
    Indexed :class:`pandas.Series` of annualized growth rates

    Raises
    ------
    ValueError
        Math domain error when timeseries crosses 0.

    See Also
    --------
    pyam.IamComputeAccessor.growth_rate

    """

    if not (all([v > 0 for v in x.values]) or all([v < 0 for v in x.values])):
        raise_data_error("Cannot compute growth rate when timeseries crosses 0", x)

    x = x.sort_index()
    growth_rate = (-x.diff(periods=-1) / x).values[:-1]  # diff on latest period is nan

    if isinstance(x.index, pd.MultiIndex):
        periods = x.index.get_level_values("year")
    else:
        periods = x.index
    period_length = -pd.Series(periods).diff(periods=-1).values[:-1]

    return pd.Series(
        [math.pow(1 + v, 1 / d) - 1 for v, d in zip(growth_rate, period_length)],
        index=x.index[:-1],
    )



import numpy as np
import pandas as pd

from pyam.exceptions import raise_data_error


def get_index_level_number(index, level):
    """Return the number of a specific level"""
    return index._get_level_number(level)


def get_index_levels(index, level):
    """Return the labels for a specific level"""

    if not isinstance(index, pd.Index):
        index = index.index  # assume that the arg `index` is a pd.DataFrame

    if isinstance(index, pd.MultiIndex):
        return list(index.levels[get_index_level_number(index, level)])

    # if index is one-dimensional, make sure that the "level" is the name
    if index.name != level:
        raise KeyError("Index does not have a level {level}")
    return list(index)


def get_index_levels_codes(df, level):
    """Return the category-values and codes for a specific level"""
    n = get_index_level_number(df.index, level)
    return df.index.levels[n], df.index.codes[n]


def get_keep_col(codes, matches):
    """Return boolean mask where *matches* appear in *codes*

    *matches* can be given as either:
    1. A boolean mask against the levels of a multiindex, or
    2. An subset of integers in *codes*
    """
    matches = np.asanyarray(matches)

    if np.issubdtype(matches.dtype, "bool"):
        (matches,) = np.where(matches)

    return np.isin(codes, matches)


def replace_index_values(df, name, mapping, rows=None):
    """Replace one or several category-values at a specific level (for specific rows)"""
    index = df if isinstance(df, pd.Index) else df.index

    n = get_index_level_number(index, name)

    # if replacing level values with a filter (by rows)
    if rows is not None and not all(rows):
        _levels = pd.Series(index.get_level_values(n))
        renamed_index = replace_index_values(index[rows], name, mapping)
        _levels[rows] = list(renamed_index.get_level_values(n))
        _unique_levels = pd.Index(_levels.unique())

        return append_index_level(
            index=index.droplevel(n),
            codes=_unique_levels.get_indexer(_levels),
            level=_unique_levels,
            name=name,
            order=index.names,
        )

    # else, replace the level values for the entire index dimension
    _levels = index.levels[n].map(lambda level: mapping.get(level, level))
    _unique_levels = _levels.unique()

    # if no duplicate levels exist after replace, set new levels and return
    if len(index.levels[n]) == len(_unique_levels):
        return index.set_levels(_levels, level=n)

    # if duplicate levels exist, re-map the codes
    level_mapping = _unique_levels.get_indexer(_levels)
    _codes = np.where(index.codes[n] != -1, level_mapping[index.codes[n]], -1)
    return index.set_codes(_codes, level=n).set_levels(_unique_levels, level=n)


def replace_index_labels(index, name, labels):
    """Replace the labels for a specific level"""

    n = get_index_level_number(index, name)
    codes = index.codes[n]
    return append_index_level(index.droplevel(n), codes, labels, name, index.names)


def append_index_col(index, values, name, order=False):
    """Append a list of `values` as a new column (level) to an `index`"""
    levels = pd.Index(values).unique()
    codes = levels.get_indexer(values)

    return append_index_level(index, codes, levels, name, order)


def append_index_level(index, codes, level, name, order=False):
    """Append a level to a pd.MultiIndex"""
    if isinstance(level, str):
        level = [level]
        codes = [codes] * len(index.codes[0])

    new_index = pd.MultiIndex(
        codes=index.codes + [codes],
        levels=index.levels + [level],
        names=index.names + [name],
    )
    if order:
        new_index = new_index.reorder_levels(order)
    return new_index


def verify_index_integrity(df):
    """Verify integrity of index

    Arguments
    ---------
    df : Union[pd.DataFrame, pd.Series, pd.Index]

    Raises
    ------
    ValueError
    """
    index = df if isinstance(df, pd.Index) else df.index
    if not index.is_unique:
        overlap = index[index.duplicated()].unique()

        raise_data_error(
            "Timeseries data has overlapping values", overlap.to_frame(index=False)
        )



import importlib.metadata
import itertools
import logging
import re
import string
import warnings
from contextlib import contextmanager
from datetime import datetime, timedelta
from pathlib import Path

import dateutil
import numpy as np
import packaging.version
import pandas as pd
from pandas.api.types import is_list_like

from pyam.exceptions import raise_data_error
from pyam.index import get_index_levels, replace_index_labels
from pyam.str import concat_with_pipe, escape_regexp, find_depth, is_str

logger = logging.getLogger(__name__)

# common indices
DEFAULT_META_INDEX = ["model", "scenario"]
META_IDX = ["model", "scenario"]
YEAR_IDX = ["model", "scenario", "region", "year"]
IAMC_IDX = ["model", "scenario", "region", "variable", "unit"]
SORT_IDX = ["model", "scenario", "variable", "year", "region"]
LONG_IDX = IAMC_IDX + ["year"]

# required columns
REQUIRED_COLS = ["region", "variable", "unit"]

# illegal terms for data/meta column names to prevent attribute conflicts
ILLEGAL_COLS = ["data", "meta", "level", "exclude", "measurand", ""]

# dictionary to translate column count to Excel column names
NUMERIC_TO_STR = dict(
    zip(
        range(0, 702),
        [i for i in string.ascii_uppercase]
        + [
            f"{i}{j}"
            for i, j in itertools.product(
                string.ascii_uppercase, string.ascii_uppercase
            )
        ],
    )
)


@contextmanager
def adjust_log_level(logger="pyam", level="ERROR"):
    """Context manager to change loglevel"""
    if isinstance(logger, str):
        logger = logging.getLogger(logger)
    old_level = logger.getEffectiveLevel()
    logger.setLevel(level)
    yield
    logger.setLevel(old_level)


def to_list(x):
    """Return x as a list"""
    return x if is_list_like(x) else [x]


def remove_from_list(x, items):
    """Remove `items` from list `x`"""
    items = to_list(items)
    return [i for i in x if i not in items]


def write_sheet(writer, name, df, index=False):
    """Write a pandas.DataFrame to an ExcelWriter

    The function applies formatting of the column width depending on maxwidth
    of values and column header

    Parameters
    ----------
    writer: pandas.ExcelWriter
        an instance of a :class:`pandas.ExcelWriter`
    name: string
        name of the sheet to be written
    df: pandas.DataFrame
        a :class:`pandas.DataFrame` to be written to the sheet
    index: boolean, default False
        should the index be written to the sheet
    """
    if index:
        df = df.reset_index()
    df.to_excel(writer, sheet_name=name, index=False)
    for i, col in enumerate(df.columns):
        if df.dtypes[col].name.startswith(("float", "int")):
            width = len(str(col)) + 2
        else:
            width = min(
                max([df[col].map(lambda x: len(str(x or "None"))).max(), len(col)]) + 2,
                80,  # make sure that column width is not too large
            )
        writer.sheets[name].set_column(i, i, width)  # assumes xlsxwriter as engine


def get_excel_file_with_kwargs(path, **kwargs):
    """Return a `pandas.ExcelFile` and a dict of unused kwargs.

    When reading an Excel file, this function finds keyword arguments that
    should be passed to `pandas.ExcelFile`, and returns a `pandas.ExcelFile`
    instance along with the remaining keyword arguments (which presumably
    will be used for other purposes by the calling function).
    """
    EXCEL_FILE_KWS = ("engine", "storage_options", "engine_kwargs")
    kwargs = kwargs.copy()
    excel_file_kwargs = {k: kwargs.pop(k) for k in EXCEL_FILE_KWS if k in kwargs}
    # TODO remove when bumping minimum pandas dependency to >= 2.2
    if "engine_kwargs" in excel_file_kwargs and packaging.version.parse(
        importlib.metadata.version("pandas")
    ) < packaging.version.parse("2.2.0"):
        warnings.warn(
            "pandas < 2.2.0 has inconsistent support for `engine_kwargs`. "
            "Using it is likely to result in an exception."
        )
    return pd.ExcelFile(path, **excel_file_kwargs), kwargs


def read_pandas(path, sheet_name=["data*", "Data*"], *args, **kwargs):
    """Read a file and return a pandas.DataFrame"""

    if isinstance(path, Path) and path.suffix == ".csv":
        return pd.read_csv(path, *args, **kwargs)

    xlfile, kwargs = get_excel_file_with_kwargs(path, **kwargs)
    with xlfile as xl:
        # reading multiple sheets
        sheet_names = pd.Series(xl.sheet_names)
        if len(sheet_names) > 1:
            sheets = kwargs.pop("sheet_name", sheet_name)
            # apply pattern-matching for sheet names (use * as wildcard)
            sheets = sheet_names[pattern_match(sheet_names, values=sheets)]
            if sheets.empty:
                raise ValueError(f"Sheet(s) '{sheet_name}' not found in file '{path}'.")

            df = pd.concat([xl.parse(s, *args, **kwargs) for s in sheets])

        # read single sheet (if only one exists in file) ignoring sheet name
        else:
            df = pd.read_excel(xl, *args, **kwargs)

    # remove unnamed and empty columns, and rows were all values are nan
    def is_empty(name, s):
        if str(name).startswith("Unnamed: "):
            try:
                if len(s) == 0 or all(np.isnan(s)):
                    return True
            except TypeError:
                pass
        return False

    empty_cols = [c for c in df.columns if is_empty(c, df[c])]
    return df.drop(columns=empty_cols).dropna(axis=0, how="all")


def read_file(path, *args, **kwargs):
    """Read data from a file"""
    # extract kwargs that are intended for `format_data`
    format_kwargs = dict(index=kwargs.pop("index"))
    for c in [i for i in IAMC_IDX + ["year", "time", "value"] if i in kwargs]:
        format_kwargs[c] = kwargs.pop(c)
    return format_data(read_pandas(path, *args, **kwargs), **format_kwargs)


def _convert_r_columns(df):
    """Check and convert R-style year columns"""

    def strip_R_integer_prefix(c):
        try:
            first = c[0]
            second = c[1:]
            if first == "X":
                try:
                    #  bingo! was X2015 R-style, return the integer
                    return int(second)
                except ValueError:
                    # nope, not an int, fall down to final return statement
                    pass
        except (TypeError, IndexError):
            # not a string/iterable/etc, fall down to final return statement
            pass
        return c

    return df.set_axis(df.columns.map(strip_R_integer_prefix), axis="columns")


def _knead_data(df, **kwargs):
    """Replace, rename and concat according to user arguments"""

    # if `value` is given but not `variable`,
    # melt value columns and use column name as `variable`
    if "value" in kwargs and "variable" not in kwargs:
        value = kwargs.pop("value")
        value = value if is_list_like(value) else [value]
        _df = df.set_index(list(set(df.columns) - set(value)))
        dfs = []
        for v in value:
            if v not in df.columns:
                raise ValueError(
                    f"Column `{v}` not in timeseries data, found: {df.columns}"
                )
            vdf = _df[v].to_frame().rename(columns={v: "value"})
            vdf["variable"] = v
            dfs.append(vdf.reset_index())
        df = pd.concat(dfs).reset_index(drop=True)

    # otherwise, rename columns or concat to IAMC-style or do a fill-by-value
    for col, value in kwargs.items():
        if col in df:
            raise ValueError(
                f"Conflict of kwarg with column `{col}` in timeseries data."
            )

        if is_str(value) and value in df:
            df.rename(columns={value: col}, inplace=True)
        elif is_list_like(value) and all([c in df.columns for c in value]):
            df[col] = df.apply(lambda x: concat_with_pipe(x, cols=value), axis=1)
            df.drop(value, axis=1, inplace=True)
        elif is_str(value):
            df[col] = value
        else:
            raise ValueError(f"Invalid argument for casting data: `{col}: {value}`")

    return df


def _format_from_legacy_database(df):
    """Process data from legacy databases (SSP and earlier)"""

    logger.info("Ignoring notes column in `data`.")
    df.drop(columns="notes", inplace=True)
    col = df.columns[0]  # first column has database copyright notice
    df = df[~df[col].str.contains("database", case=False)]
    if "scenario" in df.columns and "model" not in df.columns:
        # model and scenario are jammed together in RCP data
        parts = df["scenario"].str.split("-", n=1, expand=True)
        df = df.assign(model=parts[0].str.strip(), scenario=parts[1].str.strip())

    return df


def _intuit_column_groups(df, index, include_index=False):  # noqa: C901
    """Check and categorise columns in dataframe"""

    if include_index:
        existing_cols = pd.Index(df.index.names)
    else:
        existing_cols = pd.Index([])
    if isinstance(df, pd.Series):
        existing_cols = existing_cols.union(["value"])
    elif isinstance(df, pd.DataFrame):
        existing_cols = existing_cols.union(df.columns)

    # check that there is no unnamed column in the timeseries data
    if None in existing_cols:
        raise ValueError("Unnamed column in timeseries data: None")

    # check that there is no column in the timeseries data with reserved/illegal names
    conflict_cols = [i for i in existing_cols if i in ILLEGAL_COLS]
    if conflict_cols:
        sep = "', '"
        _cols = f"'{sep.join(conflict_cols)}'"
        _args = ", ".join([f"<alternative_column_name>='{i}'" for i in conflict_cols])
        raise ValueError(
            f"Illegal column{s(len(conflict_cols))} for timeseries data: {_cols}\n"
            f"Use `IamDataFrame(..., {_args})` to rename at initialization."
        )

    # check that index and required columns exist
    missing_index = [c for c in index if c not in existing_cols]
    if missing_index:
        raise ValueError(f"Missing index columns in timeseries data: {missing_index}")

    missing_required_col = [c for c in REQUIRED_COLS if c not in existing_cols]
    if missing_required_col:
        raise ValueError(
            f"Missing required columns in timeseries data: {missing_required_col}"
        )

    # check whether data in wide format (standard IAMC) or long format (`value` column)
    if "value" in existing_cols:
        # check if time column is given as `year` (int) or `time` (datetime)
        if "year" in existing_cols and "time" not in existing_cols:
            time_col = "year"
        elif "time" in existing_cols and "year" not in existing_cols:
            time_col = "time"
        else:
            raise ValueError("Invalid time domain, must have either `year` or `time`.")
        extra_cols = [
            c
            for c in existing_cols
            if c not in index + REQUIRED_COLS + [time_col, "value"]
        ]
        data_cols = []
    else:
        # if in wide format, check if columns are years (int) or datetime
        cols = [c for c in existing_cols if c not in index + REQUIRED_COLS]
        year_cols, time_cols, extra_cols = [], [], []
        for i in cols:
            # if the column name can be cast to integer, assume it's a year column
            try:
                int(i)
                year_cols.append(i)

            # otherwise, try casting to datetime
            except (ValueError, TypeError):
                try:
                    dateutil.parser.parse(str(i))
                    time_cols.append(i)

                # neither year nor datetime, so it is an extra-column
                except ValueError:
                    extra_cols.append(i)

        if year_cols and not time_cols:
            time_col = "year"
            data_cols = sorted(year_cols)
        else:
            time_col = "time"
            data_cols = sorted(year_cols) + sorted(time_cols)
        if not data_cols:
            raise ValueError("No time domain in the data.")

    return time_col, extra_cols, data_cols


def _format_data_to_series(df, index):
    """Convert a long or wide pandas dataframe to a series with the required columns"""

    time_col, extra_cols, data_cols = _intuit_column_groups(df, index)

    _validate_complete_index(df[index + REQUIRED_COLS + extra_cols])

    idx_order = index + REQUIRED_COLS + [time_col] + extra_cols

    if data_cols:
        # wide format
        df = (
            df.set_index(index + REQUIRED_COLS + extra_cols)
            .rename_axis(columns=time_col)
            .stack()
            .rename("value")
            .reorder_levels(idx_order)
        )
    else:
        # long format
        df = df.set_index(idx_order)["value"].dropna()

    return df, time_col, extra_cols


def _check_data(df, time_col):
    """Validate data series for infinite values, duplicates, and emptiness"""

    # check for infinite values
    inf_rows = np.isinf(df)
    if inf_rows.any():
        raise_data_error(
            "Infinite values in `data`", df[inf_rows].index.to_frame(index=False)
        )

    # check for duplicate rows
    rows = df.index.duplicated()
    if any(rows):
        raise_data_error(
            "Duplicate rows in `data`", df[rows].index.to_frame(index=False)
        )

    # check if data is empty
    if df.empty:
        logger.warning("Formatted data is empty.")

    return df


def format_data(df, index, **kwargs):  # noqa: C901
    """Convert a pandas.Dataframe or pandas.Series to the required format"""

    # Fast-pass if `df` has the index and required columns as a pd.MultiIndex
    if set(df.index.names) >= set(index) | set(REQUIRED_COLS) and not kwargs:
        time_col, extra_cols, data_cols = _intuit_column_groups(
            df, index=index, include_index=True
        )

        if isinstance(df, pd.DataFrame):
            extra_cols_not_in_index = [c for c in extra_cols if c in df.columns]
            if extra_cols_not_in_index:
                df = df.set_index(extra_cols_not_in_index, append=True)

            if data_cols:
                df = (
                    df[data_cols]
                    .rename_axis(columns=time_col)
                    .stack(future_stack=True)
                    .rename("value")
                )
            else:
                df = df["value"]

        df = df.reorder_levels(index + REQUIRED_COLS + [time_col] + extra_cols).dropna()

        # remove unused levels to guard against issue #762
        df.index = df.index.remove_unused_levels()

    else:
        if isinstance(df, pd.Series):
            if not df.name:
                df = df.rename("value")
            df = df.reset_index()
        elif not list(df.index.names) == [None]:
            # reset the index if meaningful entries are included there
            df = df.reset_index()

        df = _convert_r_columns(df)

        if kwargs:
            df = _knead_data(df, **kwargs)

        # all lower case
        df.rename(
            columns={c: str(c).lower() for c in df.columns if is_str(c)}, inplace=True
        )

        if "notes" in df.columns:  # this came from a legacy database (SSP or earlier)
            df = _format_from_legacy_database(df)

        # replace missing units by an empty string for user-friendly filtering
        if "unit" in df.columns:
            df = df.assign(unit=df["unit"].fillna(""))

        df, time_col, extra_cols = _format_data_to_series(df, index)

    # cast value column to numeric
    try:
        df = pd.to_numeric(df)
    except ValueError as e:
        # get the row number where the error happened
        row_nr_regex = re.compile(r"(?<=at position )\d+")
        row_nr = int(row_nr_regex.search(str(e)).group())
        short_error_regex = re.compile(r".*(?= at position \d*)")
        short_error = short_error_regex.search(str(e)).group()
        raise_data_error(f"{short_error} in `data`", df.iloc[[row_nr]])

    # format the time-column
    _time = [to_time(i) for i in get_index_levels(df.index, time_col)]
    df.index = replace_index_labels(df.index, time_col, _time)

    # perform data validation checks
    df = _check_data(df, time_col)

    return df, index, time_col, extra_cols


def _validate_complete_index(df):
    """Validate that there are no nan's in the (index) columns"""
    null_cells = df.isnull()
    null_rows = null_cells.any(axis=1)
    if null_rows.any():
        null_cols = null_cells.any()
        cols = ", ".join(null_cols.index[null_cols])
        raise_data_error(
            f"Empty cells in `data` (columns: '{cols}')", df.loc[null_rows]
        )


def merge_meta(left, right, ignore_conflict=False):
    """Merge two `meta` tables; raise if values are in conflict (optional)

    If conflicts are ignored, values in `left` take precedence over `right`.
    """
    left = left.copy()  # make a copy to not change the original object
    diff = right.index.difference(left.index)
    sect = right.index.intersection(left.index)

    # merge `right` into `left` for overlapping scenarios (`sect`)
    if not sect.empty:
        # if not ignored, check that overlapping `meta` columns are equal
        if not ignore_conflict:
            cols = [i for i in right.columns if i in left.columns]
            if not left.loc[sect, cols].equals(right.loc[sect, cols]):
                conflict_idx = (
                    pd.concat([right.loc[sect, cols], left.loc[sect, cols]])
                    .drop_duplicates()
                    .index.drop_duplicates()
                )
                msg = "Conflict in `meta` for scenarios {}".format(
                    [i for i in pd.DataFrame(index=conflict_idx).index]
                )
                raise ValueError(msg)
        # merge new columns
        cols = [i for i in right.columns if i not in left.columns]
        left = left.merge(
            right.loc[sect, cols], how="outer", left_index=True, right_index=True
        )

    # join `other.meta` for new scenarios (`diff`)
    if not diff.empty:
        left = pd.concat([left, right.loc[diff, :]], sort=False)

    # remove any columns that are all-nan
    return left.dropna(axis=1, how="all")


def merge_exclude(left, right, ignore_conflict=False):
    """Merge two `exclude` series; raise if values are in conflict (optional)"""

    left = left.copy()  # make a copy to not change the original object
    diff = right.index.difference(left.index)
    sect = right.index.intersection(left.index)

    # if not ignored, check that overlapping `meta` columns are equal
    if not sect.empty:
        conflict = left[sect][left[sect] != right[sect]].index
        if not conflict.empty:
            n = len(conflict)
            if ignore_conflict:
                logger.warning(f"Ignoring conflict{s(n)} in `exclude` attribute.")
            else:
                raise_data_error(
                    f"Conflict when merging `exclude` for the following scenario{s(n)}",
                    conflict,
                )
    return pd.concat([left, right.loc[diff]], sort=False)


def make_index(df, cols=META_IDX, unique=True):
    """Create an index from the columns/index of a dataframe or series"""

    def _get_col(c):
        try:
            return df.index.get_level_values(c)
        except KeyError:
            return df[c]

    index = pd.MultiIndex.from_tuples(
        list(zip(*[_get_col(col) for col in cols])), names=tuple(cols)
    )
    return index.drop_duplicates() if unique else index


def pattern_match(
    data, values, level=None, regexp=False, has_nan=False, return_codes=False
):
    """Return list where data matches values

    The function matches model/scenario names, variables, regions
    and meta columns to pseudo-regex (if `regexp == False`)
    for filtering (str, int, bool)
    """
    codes = []
    matches = np.zeros(len(data), dtype=bool)
    values = values if is_list_like(values) else [values]

    # issue (#40) with string-to-nan comparison, replace nan by empty string
    _data = data.fillna("") if has_nan else data

    for s in values:
        if return_codes and isinstance(data, pd.Index):
            try:
                codes.append(data.get_loc(s))
                continue
            except KeyError:
                pass

        if is_str(s):
            pattern = re.compile(escape_regexp(s) + "$" if not regexp else s)
            depth = True if level is None else find_depth(_data, s, level)
            matches |= data.str.match(pattern) & np.array(depth)
        else:
            matches = np.logical_or(matches, data == s)

    if return_codes:
        codes.extend(np.where(matches)[0])
        return codes

    return matches


def print_list(x, n):
    """Return a printable string of a list shortened to n characters"""
    # if list is empty, only write count
    if len(x) == 0:
        return "(0)"

    # write number of elements, subtract count added at end from line width
    x = [i if i != "" else "''" for i in map(str, x)]
    count = f" ({len(x)})"
    n -= len(count)

    # if not enough space to write first item, write shortest sensible line
    if len(x[0]) > n - 5:
        return "..." + count

    # if only one item in list
    if len(x) == 1:
        return f"{x[0]} (1)"

    # add first item
    lst = f"{x[0]}, "
    n -= len(lst)

    # if possible, add last item before number of elements
    if len(x[-1]) + 4 > n:
        return lst + "..." + count
    else:
        count = f"{x[-1]}{count}"
        n -= len({x[-1]}) + 3

    # iterate over remaining entries until line is full
    for i in x[1:-1]:
        if len(i) + 6 <= n:
            lst += f"{i}, "
            n -= len(i) + 2
        else:
            lst += "... "
            break

    return lst + count


# utility method to compare years (as integer) and datetime for index-sorting
def compare_year_time(x):
    return pd.Index(
        [
            # set year lower than first timestep of that year (2010 < 2010-01-01 00:00)
            datetime(time, 1, 1, 0, 0, 0) - timedelta(0, 0.01)
            if isinstance(time, int)
            else time
            for time in x
        ]
    )


def to_time(x):
    """Cast a value to either year (int) or datetime"""

    # if the column name can be cast to integer, assume it's a year column
    try:
        j = int(x)
        is_year = True

    # otherwise, try casting to Timestamp (pandas-equivalent of datetime)
    except (ValueError, TypeError):
        try:
            j = pd.Timestamp(x)
            is_year = False
        except ValueError:
            raise ValueError(f"Invalid time domain: {x}")

    # This is to guard against "years" with decimals (e.g., '2010.5')
    if is_year and float(x) != j:
        raise ValueError(f"Invalid time domain: {x}")

    return j


def to_int(x, index=False):
    """Formatting series or timeseries columns to int and checking validity

    If `index=False`, the function works on the :class:`pandas.Series` x;
    else, the function casts the index of x to int and returns x with new index
    """
    _x = x.index if index else x
    cols = list(map(int, _x))
    error = _x[cols != _x]
    if not error.empty:
        raise ValueError(f"Invalid values: {error}")
    if index:
        x.index = cols
        return x
    else:
        return _x


def s(n):
    """Return an s if n!=1 for nicer formatting of log messages"""
    return "s" if n != 1 else ""



import operator

import pandas as pd
from iam_units import registry
from pint import Quantity

from pyam.index import append_index_level, get_index_levels, replace_index_values
from pyam.utils import to_list


# these functions have to be defined explicitly to allow calling them with keyword args
def add(a, b):
    return operator.add(*_make_series(a, b))


def subtract(a, b):
    return operator.sub(*_make_series(a, b))


def multiply(a, b):
    return operator.mul(*_make_series(a, b))


def divide(a, b):
    return operator.truediv(*_make_series(a, b))


def _make_series(a, b):
    """Cast instances of a pint.Quantity to a pd.Series

    Calling an operation on a pd.Series and a pint.Quantity removes the units,
    therefore a pint.Quantity is transformed to a pd.Series of quantities.
    """
    if isinstance(a, Quantity):
        return pd.Series([a] * len(b), index=b.index), b
    if isinstance(b, Quantity):
        return a, pd.Series([b] * len(a), index=a.index)
    return a, b


KNOWN_OPS = {
    "add": add,
    "subtract": subtract,
    "multiply": multiply,
    "divide": divide,
}


def _op_data(df, name, method, axis, fillna=None, args=(), ignore_units=False, **kwds):  # noqa: C901
    """Internal implementation of numerical operations on timeseries"""

    if axis not in df.dimensions:
        raise ValueError(f"Unknown axis: {axis}")

    if method in KNOWN_OPS:
        method = KNOWN_OPS[method]
    elif callable(method):
        pass
    else:
        raise ValueError(f"Unknown method: {method}")

    cols = [d for d in df.dimensions if d != axis]

    # replace args and kwds with values of `df._data` if applicable
    # _data_args and _data_kwds track if an argument was replaced by `df._data` values
    n = len(args)
    _args, _data_args, _units_args = [None] * n, [False] * n, [None] * n
    for i, value in enumerate(args):
        _args[i], _units_args[i], _data_args[i] = _get_values(
            df, axis, value, cols, f"_arg{i}"
        )

    _data_kwds, _unit_kwds = {}, {}
    for key, value in kwds.items():
        kwds[key], _unit_kwds[key], _data_kwds[key] = _get_values(
            df, axis, value, cols, key
        )

    # fast-pass on units: override pint for some methods if all kwds have the same unit
    if (
        method in [add, subtract, divide]
        and ignore_units is False
        and fillna is None
        and len(_unit_kwds["a"]) == 1
        and len(_unit_kwds["b"]) == 1
        and _unit_kwds["a"][0] == _unit_kwds["b"][0]
    ):
        # activate ignore-units feature
        ignore_units = _unit_kwds["a"][0] if method in [add, subtract] else ""
        # downcast `pint.Quantity` to numerical value
        kwds["a"], kwds["b"] = _to_value(kwds["a"]), _to_value(kwds["b"])

    # cast args and kwds to pd.Series of pint.Quantity
    if ignore_units is False:
        for i, is_data in enumerate(_data_args):
            _args[i] = _to_quantity(_args[i]) if is_data else _args[i]
        for key, value in kwds.items():
            kwds[key] = _to_quantity(value) if _data_kwds[key] else value
    # else remove units from pd.Series
    else:
        for i, is_data in enumerate(_data_args):
            _args[i] = _args[i].reset_index("unit", drop=True) if is_data else _args[i]
        for key, value in kwds.items():
            kwds[key] = (
                value.reset_index("unit", drop=True) if _data_kwds[key] else value
            )

    # merge all args and kwds that are based on `df._data` to apply fillna
    if fillna is not None:
        _data_cols = [_args[i] for i, is_data in enumerate(_data_args) if is_data]
        _data_cols += [kwds[key] for key, is_data in _data_kwds.items() if is_data]
        _data = pd.merge(*_data_cols, how="outer", left_index=True, right_index=True)

        _data.fillna(fillna, inplace=True)

        for i, is_data in enumerate(_data_args):
            if is_data:
                _args[i] = _data[f"_arg{i}"]
        for key, is_data in _data_kwds.items():
            if is_data:
                kwds[key] = _data[key]

    # apply method and check that returned object is valid
    result = method(*_args, **kwds)
    if not isinstance(result, pd.Series):
        msg = f"Value returned by `{method.__name__}` cannot be cast to an IamDataFrame"
        raise ValueError(f"{msg}: {result}")

    # separate pint quantities into numerical value and unit (as index)
    if ignore_units is False:
        _value = pd.DataFrame(
            [[i.magnitude, f"{i.units:~}"] for i in result.values],
            columns=["value", "unit"],
            index=result.index,
        ).set_index("unit", append=True)
        _value.index = replace_index_values(_value, "unit", {"dimensionless": ""})

    # otherwise, set unit (as index) to "unknown" or the value given by "ignore_units"
    else:
        index = append_index_level(
            result.index,
            codes=0,
            level="unknown" if ignore_units is True else ignore_units,
            name="unit",
        )
        _value = pd.Series(result.values, index=index, name="value")

    # append the `name` to the index on the `axis`
    _value.index = append_index_level(_value.index, codes=0, level=name, name=axis)
    return _value


def _get_values(df, axis, value, cols, name):
    """Return grouped data if value is in axis. Otherwise return value.

    Parameters
    ----------
    df : IamDataFrame
        IamDataFrame to select the values from.
    axis : str
        Axis in `df` that contains value.
    value : str or list of str or any
        Either str or list of str in axis or anything else.
    cols : list
        Columns in df that are not `axis`.
    name : str
        Name of the returned pd.Series.

    Returns
    -------
    Tuple of the following:
     - Either `df.data` downselected by `{axis: value}` or `value`
     - List of units of the timeseries data or `value`
     - Bool whether first item was derived from `df.data`

    """
    # check if `value` is a `pint.Quantity` and return unit specifically
    if isinstance(value, Quantity):
        return value, [value.units], False
    # try selecting from `df.data`
    if any(v in get_index_levels(df._data, axis) for v in to_list(value)):
        _df = df.filter(**{axis: value})
        return _df._data.groupby(cols).sum().rename(index=name), _df.unit, True
    # else, return value
    return value, [], False


def _to_quantity(data):
    """Convert the values of an indexed pd.Series into pint.Quantity instances"""
    return pd.Series(
        [
            registry.Quantity(v, u)
            for v, u in zip(data.values, data.index.get_level_values("unit"))
        ],
        index=data.reset_index("unit", drop=True).index,
        name=data.name,
    )


def _to_value(x):
    """Return the value of a pint.Quantity"""
    if isinstance(x, Quantity):
        return x.magnitude
    return x



import datetime as dt

import numpy as np
import pandas as pd

from pyam.index import get_index_levels

try:
    import xarray as xr

    HAS_XARRAY = True
except ModuleNotFoundError:
    xr = None
    HAS_XARRAY = False
from pyam.utils import IAMC_IDX, META_IDX

NETCDF_IDX = ["time", "model", "scenario", "region"]


def read_netcdf(path):
    """Read timeseries data and meta indicators from a netCDF file

    Parameters
    ----------
    path : :class:`pathlib.Path` or file-like object
        Scenario data file in netCDF format.

    Returns
    -------
    :class:`IamDataFrame`

    See Also
    --------
    pyam.IamDataFrame.to_netcdf

    Notes
    -----
    Read the `pyam-netcdf docs <https://pyam-iamc.readthedocs.io/en/stable/api/io.html>`_
    for more information on the expected file format structure.

    """
    from pyam import IamDataFrame

    if not HAS_XARRAY:
        raise ModuleNotFoundError("Reading netcdf files requires 'xarray'.")
    _ds = xr.open_dataset(path)
    _list_variables = [i for i in _ds.to_dict()["data_vars"].keys()]

    # Check if the time coordinate is years (integers) or date time-format
    is_year_based = all(
        isinstance(x, (int, np.integer)) for x in _ds.coords["time"].values
    )
    is_datetime = all(
        isinstance(x, (dt.date, dt.time, np.datetime64))
        for x in _ds.coords["time"].values
    )

    # Check if the xarray dataset has the correct coordinates, then get column names
    if is_year_based:
        _list_cols = IAMC_IDX + ["year", "value"]
    elif is_datetime:
        _list_cols = IAMC_IDX + ["time", "value"]
    else:
        raise TypeError(
            "Time coordinates can year (integer) or datetime format, found: "
            + _ds.coords["time"]
        )

    # read `data` table
    _data = []
    _meta = []
    for _var in _list_variables:
        # Check dimensions, if exactly as in META_IDX is a meta indicator
        # if exactly as in IAMC_IDX is a variable
        if set(_ds[_var].dims) == set(META_IDX):
            _meta.append(_var)
        elif set(_ds[_var].dims) == set(NETCDF_IDX):
            # convert the data into the IamDataframe format
            _tmp = (
                _ds[_var]
                .to_dataframe()
                .rename(columns={_var: "value"})
                .reset_index(drop=False)
            )
            _tmp["variable"] = _ds[_var].long_name
            _tmp["unit"] = _ds[_var].unit
            _data.append(_tmp)
        else:
            raise TypeError(
                f"Cannot define {_var}, different indices from META_IDX and IAMC_IDX."
            )
    data = pd.concat(_data).reset_index(drop=True)
    # if year-based data, get the time coordinate as "year"
    # if timeseries, keep the time coordinate as "time"
    if is_year_based:
        data = data.rename(columns={"time": "year"})

    return IamDataFrame(
        data,
        meta=_ds[_meta].to_dataframe().replace("nan", np.nan) if _meta else None,
    )


def to_xarray(data_series: pd.Series, meta: pd.DataFrame):
    """Convert timeseries data and meta indicators to an xarray Dataset

    Returns
    -------
    :class:`xarray.Dataset`

    """
    if not HAS_XARRAY:
        raise ModuleNotFoundError("Converting to xarray requires 'xarray'.")

    dataset = xr.Dataset()

    # add timeseries data-variables
    for variable, _variable_data in data_series.groupby("variable"):
        unit = get_index_levels(_variable_data, "unit")

        if len(unit) > 1:
            raise ValueError(
                "Cannot write to xarray for non-unique units in '" + variable + "'."
            )

        dataset[variable] = xr.DataArray(
            _variable_data.droplevel(["variable", "unit"]).to_xarray(),
        )
        dataset[variable].attrs = {
            "unit": unit[0],
            "long_name": variable,
        }

    # add meta indicators as data-variables
    for meta_indicator, meta_data in meta.items():
        meta_data = meta_data.replace(np.nan, "nan")
        dataset[meta_indicator] = xr.DataArray(
            meta_data.to_xarray(),
            dims=META_IDX,
            name=meta_indicator,
        )

    return dataset
